// -------------------------------------------------------------------
// Block.java:
//    This class represents a block of htmlpx code.  The block can be
//    of three types:
//       1.  A function
//       2.  A loop
//       3.  A definition
// -------------------------------------------------------------------
import java.io.*;
import java.util.*;

public class Block {
   int      iType;
   int      iLast;
   Vector   vLines;

   // constructor, plain
   public Block() {
      iLast = -1;
      iType = 3;        // default to a definition
      vLines = new Vector();
   }

   public Block(int type) {
      iLast = -1;
      iType = 3;        // default to a definition
      if (type > 0 && type < 4)
         iType = type;
      vLines = new Vector();
   }

   public int getType() {
      return iType;
   }

   public void add(String aline) {
      // there is nothing to add if the line is blank, skip it

      if (aline.length() > 0)
         vLines.addElement(aline);
   }

   public String getFirst() {
      iLast = 1;
      if (vLines.size() > 0)
         return (String) vLines.firstElement();
      return "";
   }

   public String getNext() {
      if (iLast > 0 && iLast < vLines.size()) {
         return (String) vLines.elementAt(iLast++);
      }
      return "";
   }

   public int load(BufferedReader rd, String endblk) {
      String buffer, work;
      StringTokenizer st;
      int cnt=0;

      while (true) {
         try {
            buffer = rd.readLine().trim();
            cnt++;
            if (buffer.length() > 0 && buffer.charAt(0) != '*') {
               // ignoring blank lines and comment lines
               st = new StringTokenizer(buffer);
               work = st.nextToken();
               if (work.equalsIgnoreCase(endblk)) {
                  // stop when we hit the end
                  break;
               }
               else
                  add(buffer);
            }
         }
         catch (Exception e) {
            break;
         }
      }

      return cnt;
   }
}
