// this class handles translating keywords to their proper values
import java.util.StringTokenizer;

public class Transltr {
   Defines defs;  // reference to the defines

   public Transltr(Defines d) {
      defs = d;
   }

   public boolean needTranslation(String inStr, String begdelim) {
      // this method will return a true if there is a keyword in the
      // string, otherwise a false
      if (inStr.indexOf(begdelim) > -1)
         return true;

      return false;
   }

   public String translateString(String inStr, String begdelim, String enddelim) {
      // this method will return the string that includes the values for
      // all keywords, if there are no keywords it will return the same
      // string
      if (!needTranslation(inStr, begdelim))
         return inStr;

      // do the translation
      int pos, pos2;
      String work;
      String buffer;
      Object newVal;
      String key;

      work = inStr;

      pos = work.indexOf(begdelim);
      while (pos > -1) {
         pos2 = work.indexOf(enddelim, pos + begdelim.length());     // find the end delimiter
         if (pos2 > 0) {
            key = work.substring(pos + begdelim.length(), pos2);
            newVal = defs.getValue(key);
            if (newVal != null) {
               String classtype = newVal.getClass().getName();
               if (classtype.indexOf("String") >= 0) {
                  // the result was a string
                  buffer = work.substring(0, pos) + newVal + work.substring(pos2 + enddelim.length());
                  work = buffer;
                  pos = work.indexOf(begdelim);
               }
               else if (classtype.indexOf("Block") >= 0) {
                  // the result was a block, handle the block
                  buffer = work.substring(0, pos) + blockToString((Block) newVal, begdelim, enddelim)
                                    + work.substring(pos2 + enddelim.length());
                  work = buffer;
                  pos = work.indexOf(begdelim);
               }
               else {
                  // we don't know how to translate it
                  pos = work.indexOf(begdelim, pos2 + enddelim.length());
               }
            }
            else
               pos = work.indexOf(begdelim, pos2 + enddelim.length());
         }
         else
            // get out of the loop because we did not find the end delimiter
            pos = -1;
      }

      return work;
   }

   // blockToString():
   //    This function will handle a block of code and convert it to a single string if it is
   //    possible to do so.  If it is a block that needs to be processed and discarded, that
   //    will be done also.
   private String blockToString(Block blk, String bd, String ed) {
      String buffer = new String("");
      String temp, token, key, val, work;
      StringTokenizer stt;
      boolean bProcess=true, bVal;
      int cond=0;

      temp = blk.getFirst();
      while (temp.length() > 0) {
         stt = new StringTokenizer(temp);
         work = stt.nextToken();    // get the first token
         if (work.equalsIgnoreCase("#IF") || work.equalsIgnoreCase("#IFDEF")) {
            bProcess = false;
            if (stt.hasMoreTokens()) {
               key = stt.nextToken().toUpperCase();
               bVal = false;
               val = new String("");   // set the default value to none
               while (stt.hasMoreTokens()) {
                  bVal = true;
                  work = stt.nextToken();
                  if (val.length() > 0)
                     val = val + " " + work;
                  else
                     val = work;
               }
               if (bVal) {
                  if (defs.isKeywordDefined(key)) {
                     if (val.compareTo((String) defs.getValue(key)) == 0)
                        cond = 1;         // do if
                     else
                        cond = 2;         // do else
                  }
                  else
                     cond = 2;            // do else
               }
               else {
                  if (defs.isKeywordDefined(key))
                     cond = 1;            // do if
                  else
                     cond = 2;            // do else
               }
            }
            else
               // if no condition, do the #IF
               cond = 1;

            if (cond == 1)
               bProcess = true;
         }
         else if (work.equalsIgnoreCase("#ELSE")) {
            if (cond == 1)
               bProcess = false;
            else
               bProcess = true;
         }
         else if (work.equalsIgnoreCase("#ENDIF")) {
            cond = 0;
            bProcess = true;
         }
         else if (bProcess) {
            // we are processing these lines, handle as normal
            if (buffer.length() > 0)
               buffer += " ";
            buffer += translateString(temp, bd, ed);
         }
         temp = blk.getNext();
      }

      return buffer;
   }
}
