// ----------------------------------------------------------------------
// htmlpx.java
//
// This application is an HTML-preprocessor that can be used to generate
// website from HTML template files.
//
// Usage:  htmlpx <sitefilename>
//
// ----------------------------------------------------------------------

import java.io.*;
import java.util.StringTokenizer;
import java.util.Stack;

public class htmlpx {
   Defines globalDefines;
   Transltr trans;
   String curInFile;
   String outputDirectory;
   String inputDirectory;
   String begdelim;
   String enddelim;
   String parmdelim;
   boolean makeall;
   long    makedate;

   // this is a NO-NO, we are using this to record the number of lines handled
   // during a define, function, forall loop
   int tempLineCnt;

   public static void main(String[] args) {
      htmlpx p = new htmlpx();

      if (args.length < 1) {
         System.out.println("Usage is:  htmlpx [-a] <site file name> [optional parameters]");
         p.printMarker();
         return;
      }

      // initialize the globalDefines table
      p.globalDefines = new Defines();
      p.trans = new Transltr(p.globalDefines);
      p.outputDirectory = new String("");
      p.inputDirectory = new String("");
      p.begdelim = new String("{");
      p.enddelim = new String("}");
      p.parmdelim = new String("|");
      p.makeall = false;
      p.makedate = 0;

      // the marker
      p.printMarker();

      // put any command-line parameters into the defines
      int firstopt = 1;     // default first optional parameter
      if (args.length > 1 && (args[0].equalsIgnoreCase("-a"))) {
         firstopt = 2;
         p.makeall = true;
         System.out.println("We are going to makeall.");
      }

      if (args.length > firstopt) {
         p.globalDefines.push();
         for (int i=firstopt; i < args.length; i++) {
            p.globalDefines.addKeyword("#CL" + i, args[i]);
         }
      }

      p.procFile(args[firstopt - 1]);
   }

   private void printMarker() {
      // the marker
      System.out.println("");
      System.out.println("----------------------------------------------------");
      System.out.println("| htmlPX 1.0 is Copyright 1998-99 by Keith Wilson. |");
      System.out.println("|                                                  |");
      System.out.println("| Visit us at http://www.wiserve.com/htmlpx.       |");
      System.out.println("----------------------------------------------------");
      System.out.println("");
   }

   private void procFile(String inFile)
   {
      // this method handles the situation where there is an input file only, it
      // passes an empty string as the output file
      procFile(inFile, "");
   }

   // ========================================================
   // this is the main file processing function
   // ========================================================
   private void procFile(String inFile, String outFile)
   {
      BufferedReader is;
      BufferedWriter os=null;

      try {
         is = new BufferedReader(new FileReader(inFile));
      }
      catch (Exception e) {
         System.out.println("Error opening site file:  " + e.toString());
         return;
      }

      // determine if there is an output file
      curInFile = inFile;

      if (outFile.length() > 0) {
         try {
            os = new BufferedWriter(new FileWriter(outFile));
         }
         catch (Exception e) {
            System.out.println("Error opening output file:  " + e.toString());
            return;
         }
      }

      insert(is, os);

      // close up the input file
      try {
         is.close();
      }
      catch (Exception e) {
         System.out.println("Error closing site file:  " + e.toString());
      }

      if (os != null) {
         try {
            os.close();
         }
         catch (Exception e) {
            System.out.println("Error closing output file:  " + e.toString());
         }
      }

      return;
   }

   // -------------------------------------------------------------
   // insert():
   //    This method takes an input reader and writes the results
   //    to the output writer.  This is the main workhorse function.
   // -------------------------------------------------------------
   private void insert(BufferedReader rd, BufferedWriter wr) {
      String buffer;
      long tempLong;
      StringTokenizer strTok;
      String work, temp, key;
      boolean lineError=false;
      int linecnt=0, outLine=0;
      int condState=0;     // 0 - nothing, 1 - handle till #else, 2 - skip till #else, 3 - skip till #end,
                           // 4 - skip till #endskip

      // initialize some variables so the compiler does not complain
      key = new String("");

      // we need a new set of properties
      globalDefines.push();

      while (true) {
         try {
            buffer = rd.readLine().trim();
            linecnt++;
            lineError = false;
            if (buffer.length() > 0 && buffer.charAt(0) != '*') {
               // ignoring blank lines and comment lines
               strTok = new StringTokenizer(buffer);
               work = strTok.nextToken();    // get the first token
               if (work.equalsIgnoreCase("#ELSE")) {
                  switch (condState) {
                     case 0:
                     case 3:
                        lineError = true;    // it shouldn't be here
                        break;

                     case 1:
                        condState = 3;       // we need to start skipping
                        break;

                     case 2:
                        condState = 0;       // start processing it now
                        break;

                     default:
                        break;
                  }
               }
               else if (work.equalsIgnoreCase("#ENDIF")) {
                  condState = 0;    // start handling as normal again
               }
               else if (condState == 4 && work.equalsIgnoreCase("#ENDSKIP")) {
                  condState = 0;    // start handling as normal again
               }
               else if (work.equalsIgnoreCase("#IF") || work.equalsIgnoreCase("#IFDEF")) {
                  boolean bVal=false;

                  if (condState == 0) {
                     // determine whether to skip or not
                     if (strTok.hasMoreTokens()) {
                        key = strTok.nextToken().toUpperCase();
                        temp = new String("");  // if there is no value, it is an empty string
                        while (strTok.hasMoreTokens()) {
                           bVal = true;
                           work = strTok.nextToken();
                           if (temp.length() > 0)
                              temp = temp + " " + work;
                           else
                              temp = work;
                        }
                        if (bVal) {
                           if (globalDefines.isKeywordDefined(key)) {
                              if (temp.compareTo((String) globalDefines.getValue(key)) == 0)
                                 condState = 1;       // do if
                              else
                                 condState = 2;       // do else
                           }
                           else
                              condState = 2;          // do else
                        }
                        else {
                           if (globalDefines.isKeywordDefined(key))
                              condState = 1;       // do if
                           else
                              condState = 2;       // do else
                        }
                     }
                     else
                        lineError = true;
                  }
                  else
                      // we do not allow nested conditionals at this time
                     lineError = true;
               }
               else if (condState > 1) {
                  // we need to skip
                  ;
               }
               else if (work.equalsIgnoreCase("#QUIT")) {
                  // they want to stop
                  System.out.println("Ending prematurely based on #QUIT statement.");
                  break;
               }
               else if (work.equalsIgnoreCase("#SKIP")) {
                  // they want to skip some stuff
                  condState = 4;
               }
               else if (work.equalsIgnoreCase("#SET")) {
                  lineError = handleSet(strTok);
               }
               else if (work.equalsIgnoreCase("#ECHO")) {
                  System.out.println(buffer.substring(6).trim());
               }
               else if (work.equalsIgnoreCase("#BANNER")) {
                  int color=0;
                  if (strTok.hasMoreTokens()) {
                     temp = strTok.nextToken();
                     if (temp.equalsIgnoreCase("GREEN")) {
                        color = 2;
                     }
                     else if (temp.equalsIgnoreCase("YELLOW")) {
                        color = 0;  // default color
                     }
                     else if (temp.equalsIgnoreCase("WHITE")) {
                        color = 1;
                     }
                     else if (temp.equalsIgnoreCase("RED")) {
                        color = 3;
                     }
                     else if (temp.equalsIgnoreCase("BLUE")) {
                        color = 4;
                     }
                     else if (temp.equalsIgnoreCase("TEXT")) {
                        color = -1;
                     }
                  }

                  writeBanner(wr, color);
               }
               else if (work.equalsIgnoreCase("#OUTDIR")) {
                  if (strTok.hasMoreTokens()) {
                     outputDirectory = strTok.nextToken();
                  }
                  else
                     outputDirectory = "";
               }
               else if (work.equalsIgnoreCase("#INDIR")) {
                  if (strTok.hasMoreTokens()) {
                     inputDirectory = strTok.nextToken();
                  }
                  else
                     inputDirectory = "";
               }
               else if (work.equalsIgnoreCase("#UNDEF")) {
                  // remove the current definition of a keyword
                  if (strTok.hasMoreTokens()) {
                     key = strTok.nextToken().toUpperCase();
                     globalDefines.removeKeyword(key);
                  }
                  else {
                     lineError = true;
                  }
               }
               else if (work.equalsIgnoreCase("#MAKE")) {
                  String in, out;
                  if (strTok.hasMoreTokens()) {
                     in = strTok.nextToken();
                     if (strTok.hasMoreTokens()) {
                        out = strTok.nextToken();

                        // determine if we should make this file
                        File f = new File(inputDirectory + in);
                        long infiledate = f.lastModified();
                        f = new File(outputDirectory + out);
                        long outfiledate = f.lastModified();
                        if (makeall || (outfiledate < makedate) || (makedate == 0) || (outfiledate < infiledate)) {
                           // process this input file
                           procFile(inputDirectory + in, outputDirectory + out);
                           System.out.println("Making... " + f);
                        }
                     }
                     else
                        lineError = true;
                  }
                  else
                     lineError = true;
               }
               else if (work.equalsIgnoreCase("#INCLUDE")) {
                  // include the specified file here
                  if (strTok.hasMoreTokens()) {
                     BufferedReader br2;

                     try {
                        br2 = new BufferedReader(new FileReader(inputDirectory + strTok.nextToken()));
                        insert(br2, wr);     // include this file
                        try {
                           br2.close();
                        }
                        catch (Exception e) {
                           System.out.println("Error closing the include file:  " + e.toString());
                        }
                     }
                     catch (Exception e) {
                        System.out.println("Error opening include file:  " + e.toString());
                     }
                  }
               }
               else if (work.equalsIgnoreCase("#PARMDELIM")) {
                  // set the delimiter for use
                  work = strTok.nextToken();
                  if (work.length() > 0)
                     parmdelim = work;
                  else
                     // set it back to the original value
                     begdelim = "|";
               }
               else if (work.equalsIgnoreCase("#BEGDELIM")) {
                  // set the delimiter for use
                  work = strTok.nextToken();
                  if (work.length() > 0)
                     begdelim = work;
                  else
                     // set it back to the original value
                     begdelim = "{";
               }
               else if (work.equalsIgnoreCase("#ENDDELIM")) {
                  // set the delimiter for use
                  work = strTok.nextToken();
                  if (work.length() > 0)
                     enddelim = work;
                  else
                     // set it back to the original value
                     enddelim = "}";
               }
               else if (work.equalsIgnoreCase("#DEFINE")) {
                  lineError = handleDefine(strTok, rd, 3);
                  linecnt += tempLineCnt;
               }
               else if (work.equalsIgnoreCase("#FORALL")) {
                  // forall has 2 parms:  the input file name and the field delimiter
                  // we are starting a block define
                  if (strTok.hasMoreTokens()) {
                     String thisDelim;
                     String inFile = strTok.nextToken();
                     if (strTok.hasMoreTokens())
                        thisDelim = strTok.nextToken();
                     else
                        thisDelim = "|";

                     // handle the forall
                     Block workBlock = new Block(2);
                     linecnt += workBlock.load(rd, "#ENDLOOP");
                     lineError = handleForAll(workBlock, inFile, thisDelim, wr, globalDefines);
                  }
                  else {
                     lineError = true;
                  }
               }
               else if (work.equalsIgnoreCase("#FUNCTION")) {
                  lineError = handleDefine(strTok, rd, 1);
                  linecnt += tempLineCnt;
               }
               else if (work.equalsIgnoreCase("#CALL")) {
                  // they are calling a function
                  lineError = handleCall(strTok, wr);
                  linecnt += tempLineCnt;
               }
               else if (work.equalsIgnoreCase("#LOAD")) {
                  // we are going to load defs from a file
                  if (strTok.hasMoreTokens()) {
                     // save the date for the makefile
                     String fn = strTok.nextToken();
                     File f = new File(inputDirectory + fn);
                     tempLong = f.lastModified();
                     if (makedate == 0 || (tempLong > makedate && tempLong > 0))
                        makedate = tempLong;
                     load(fn);
                  }
                  else
                     lineError = true;
               }
               else if (work.equalsIgnoreCase("#USING")) {
                  // we are going to load defs from a file
                  if (strTok.hasMoreTokens()) {
                     // save the date for the makefile
                     String fn = strTok.nextToken();
                     File f = new File(inputDirectory + fn);
                     tempLong = f.lastModified();
                     if (makedate == 0 || (tempLong > makedate && tempLong > 0))
                        makedate = tempLong;
                  }
                  else
                     lineError = true;
               }
               else {
                  // it must be a regular line, we need to translate any keywords
                  if (trans.needTranslation(buffer, begdelim))
                     temp = trans.translateString(buffer, begdelim, enddelim);
                  else
                     temp = buffer;

                  if (temp.length() > 0) {
                     if (wr != null) {
                        outLine++;
                        wr.write(temp, 0, temp.length());
                        wr.newLine();
                     }
                     else
                        System.out.println(temp);
                  }
               }

               // if there was an error, display it
               if (lineError) {
                  System.out.println("(" + curInFile + ") Error in line " + linecnt + ":  " + buffer);
               }
            }
            else if (buffer.length() == 0) {
               // NO NEED TO INCLUDE BLANK LINES!

               // we need to include the blank lines in the output file
               //if (wr != null) {
               //   outLine++;
               //   wr.newLine();
               //}
               //else
               //   System.out.println("");
            }
         }
         catch (Exception e) {
            //System.out.println("Error=" + e.toString());
            break;
         }
      }

      // we are done with this set of defines, get rid of them
      globalDefines.pop();

      return;
   }

   // ---------------------------------------------------------------------------------
   // load():
   //    This function will load defines, sets, functions, and loops from a load file
   // ---------------------------------------------------------------------------------
   private void load(String inputFile) {
      BufferedReader rd;
      String buffer;
      StringTokenizer strTok;
      String work, temp, key;
      boolean lineError;
      int linecnt=0;

      try {
         rd = new BufferedReader(new FileReader(inputDirectory + inputFile));

         while (true) {
            try {
               buffer = rd.readLine().trim();
               linecnt++;
               lineError = false;
               if (buffer.length() > 0 && buffer.charAt(0) != '*') {
                  // ignoring blank lines and comment lines
                  strTok = new StringTokenizer(buffer);
                  work = strTok.nextToken();    // get the first token
                  if (work.equalsIgnoreCase("#SET")) {
                     lineError = handleSet(strTok);
                  }
                  else if (work.equalsIgnoreCase("#ECHO")) {
                     System.out.println(buffer.substring(6).trim());
                  }
                  else if (work.equalsIgnoreCase("#DEFINE")) {
                     lineError = handleDefine(strTok, rd, 3);
                     linecnt += tempLineCnt;
                  }
                  else if (work.equalsIgnoreCase("#FUNCTION")) {
                     lineError = handleDefine(strTok, rd, 1);
                     linecnt += tempLineCnt;
                  }
                  else
                     lineError = true;   // we don't know what to do with this one

                  // if there was an error, display it
                  if (lineError) {
                     System.out.println("(" + inputFile + ") Error in line " + linecnt + ":  " + buffer);
                  }
               }
            }
            catch (Exception e) {
               //System.out.println("Error=" + e.toString());
               break;
            }
         }

         // close up the loaded file
         try {
            rd.close();
         }
         catch (Exception e) {
            System.out.println("Error closing the load file:  " + e.toString());
         }
      }
      catch (Exception e) {
         System.out.println("Error opening load file:  " + e.toString());
      }

      return;
   }

   private void writeBanner(BufferedWriter wr, int color) {
      String border, bg, text;

      switch (color) {
         case 1:  // white
            border = new String("black");
            bg = new String("#ffffff");
            text = new String("red");
            break;

         case 2:  // green
            border = new String("yellow");
            bg = new String("#408080");
            text = new String("#000000");
            break;

         case 3:  // red
            border = new String("black");
            bg = new String("red");
            text = new String("white");
            break;

         case 4:  // blue
            border = new String("red");
            bg = new String("blue");
            text = new String("white");
            break;

         case -1:  // text
            // just put the "banner" out as text
            try {
               wr.write("<font size=2>This site generated by <a href=\"");
               wr.write("http://wiserve.com/kwilson/htmlpx.htm\"><i>htmlPX.</i></a></font>");
               wr.newLine();
            }
            catch (Exception e) {
               ;
            }
            return;

         default: // yellow, default
            border = new String("blue");
            bg = new String("yellow");
            text = new String("red");
            break;
      }

      try {
         wr.write("<table border=1 bgcolor=\"" + border + "\"><tr><td bgcolor=\"" + bg + "\">");
         wr.newLine();
         wr.write("<a href=\"http://wiserve.com/kwilson/htmlpx.htm\">");
         wr.newLine();
         wr.write("<i><font size=2>This website was generated by ");
         wr.newLine();
         wr.write("<b></font><font color=\"" + text + "\" size=3>htmlPX</font></b></a></i>.</td></tr></table>");
         wr.newLine();
      }
      catch (Exception e) {
         System.out.println("Error writing banner.");
      }
   }

   //----------------------------------------------------------------------
   // handle the duplicated commands in functions
   //----------------------------------------------------------------------
   private boolean handleSet(StringTokenizer strTok) {
      // save the definition of a keyword
      if (strTok.hasMoreTokens()) {
         String key, temp, work;

         key = strTok.nextToken().toUpperCase();
         temp = new String("");  // if there is no value, it is an empty string
         while (strTok.hasMoreTokens()) {
            work = strTok.nextToken();
            if (temp.length() > 0)
               temp = temp + " " + work;
            else
               temp = work;
         }
         //System.out.println("Key/value:  " + key + "/" + temp);
         globalDefines.addKeyword(key, temp);
      }
      else
         return true;

      return false;     // no error
   }

   private boolean handleCall(StringTokenizer strTok, BufferedWriter wr) {
      // prepare a new set of variables for the parms, $1 - $n, also $NP
      if (strTok.hasMoreTokens()) {
         Integer num= new Integer(0);
         // get the function name to call
         String func = strTok.nextToken().toUpperCase();

         // we need a new set of properties
         globalDefines.push();

         int i=0;
         String buffer = new String("");
         while (strTok.hasMoreTokens()) {
            if (i > 0)
               buffer += " ";
            buffer += strTok.nextToken();
            i++;
         }

         if (i > 0) {
            num = parseParms(buffer, globalDefines, parmdelim);
         }

         globalDefines.addKeyword("$NP", num.toString());

         // process the block of code now
         String temp = trans.translateString(begdelim + func + enddelim, begdelim, enddelim);

         if (temp.length() > 0) {
            if (wr != null) {
               try {
                  wr.write(temp, 0, temp.length());
                  wr.newLine();
               }
               catch (Exception e) {
                  ;
               }
               //outLine++;
            }
            else
               System.out.println(temp);
         }

         // get rid of these parameters
         globalDefines.pop();
      }
      else
         // we don't have a define/function to call
         return true;

      return false;     // no error
   }

   private boolean handleForAll(Block blk, String inFile, String delim, BufferedWriter wr, Defines defns) {
      boolean rc=false;
      BufferedReader is;
      String buffer, temp;
      Integer parmcnt = new Integer(0);

      // open and begin reading the input file
      try {
         is = new BufferedReader(new FileReader(inFile));
      }
      catch (Exception e) {
         System.out.println("Error opening database file:  " + e.toString());
         return true;      // there was an error
      }

      while (true) {
         try {
            buffer = is.readLine();
            if (buffer.length() > 0) {
               if (buffer.substring(0, 6).equalsIgnoreCase("#STOP#"))
                  break;

               defns.push();     // create a new set of variables
               defns.addKeyword("FUNC", blk);

               parmcnt = parseParms(buffer, defns, delim);
               defns.addKeyword("$NP", parmcnt.toString());
               temp = trans.translateString(begdelim + "FUNC" + enddelim, begdelim, enddelim);
               if (temp.length() > 0) {
                  if (wr != null) {
                     //outLine++;
                     wr.write(temp, 0, temp.length());
                     wr.newLine();
                  }
                  else
                     System.out.println(temp);
               }

               defns.pop();
            }
         }
         catch (Exception e) {
            break;      // we must have hit the EOF
         }
      }

      // clean up and get out
      try {
         is.close();
      }
      catch (Exception e) {
         ;
      }

      return rc;
   }

   private Integer parseParms(String parms, Defines defns, String delim) {
      int i=0;
      String work, key, val;
      int ndx;

      work = new String(parms);
      ndx = work.indexOf(delim);
      while (ndx > 0) {
         i++;
         key = new String("$" + i);
         val = new String(work.substring(0, ndx));
         defns.addKeyword(key, val);
         work = work.substring(ndx + 1);
         ndx = work.indexOf(delim);
      }

      // get the last parameter
      if (work.length() > 0) {
         i++;
         key = new String("$" + i);
         defns.addKeyword(key, work);
      }

      Integer retInt = new Integer(i);
      return retInt;
   }

   private boolean handleDefine(StringTokenizer strTok, BufferedReader rd, int typ) {
      // we are starting a block define
      if (strTok.hasMoreTokens()) {
         String key;

         Block workBlock = new Block(typ);
         key = strTok.nextToken().toUpperCase();
         tempLineCnt = workBlock.load(rd, "#END");
         globalDefines.addKeyword(key, workBlock);
         return false;
      }

      return true;
   }
}

